import abc
import numpy as np


class Algorithm(metaclass=abc.ABCMeta):

    def __init__(self, **kwargs):
        self.population = kwargs.pop('population', 80)  # population number
        self.func = kwargs.pop('func', Ackley())  # function name
        self.max_fes = kwargs.pop('max_fes', 5e4)  # costfunction maximum number of calls
        self.Rand = np.random.RandomState(kwargs.pop('seed', 1))  # Random seed

        self.lower, self.upper = np.asarray(self.func.min_values), np.asarray(
            self.func.max_values)  # Upper and lower bounds of the function
        self.dim = self.func.dimension  # problem dimension
        self.eval_counter = 0  # Number of calls to the costfunction

        self.best_solution = list()  # save the best solutions
        self.current_solution = list()  # save the best solution in each generation

    # generate initial population
    def initial_position(self):
        return self.Rand.uniform(np.asarray(self.func.min_values), np.asarray(self.func.max_values),
                                 [self.population, self.func.dimension])

    # Coordinate boundary processing
    def boundary_handle(self, x):
        ir = np.where(x < self.lower)
        x[ir] = self.lower[ir]
        ir = np.where(x > self.upper)
        x[ir] = self.upper[ir]
        return x

    # Costfunction
    def costfunction(self, position):
        self.eval_counter += 1
        return self.func.eval(position)

    # stopping criteria
    def stopping_criteria(self):
        return self.eval_counter >= self.max_fes

    @staticmethod
    def run(self):
        pass


class Ackley:

    def __init__(self, min_values=[-32.768] * 2, max_values=[32.768] * 2, dimension=2):
        self.min_values = min_values
        self.max_values = max_values
        self.dimension = dimension

    def get_optimum(self):
        return [0] * self.dimension, 0

    @staticmethod
    def eval(sol):
        a = 20
        b = 0.2
        c = 2 * np.pi
        d = len(sol)

        sum1 = 0.0
        sum2 = 0.0
        for i in range(d):
            xi = sol[i]
            sum1 += xi ** 2
            sum2 += np.cos(c * xi)
        part1 = -a * np.exp(-b * np.sqrt(sum1 / d))
        part2 = -np.exp(sum2 / d)

        return part1 + part2 + a + np.exp(1)
