import numpy as np
import pandas as pd
from Algorithms.algorithm import Algorithm


class FS_niching:

    def __init__(self, position, niching_size, benchmark):
        """
        :param position: The population
        :param niching_size: cluster size
        :param benchmark: benchmark function
        """
        self.position = position
        self.niching_size = niching_size
        self.benchmark = benchmark

    def FS_Niching(self):
        referpoint = np.random.uniform(np.asarray(self.benchmark.min_values), np.asarray(self.benchmark.max_values),
                                       [1, self.benchmark.dimension])  # Reference point
        niche = []
        if len(self.position) % self.niching_size == 0:
            sorted_ant = self.crowddistance(referpoint, self.position)
            for i in range(int(len(self.position) / self.niching_size)):
                if len(sorted_ant) <= self.niching_size:
                    subniche = self.nearestdistance(sorted_ant[0, :], sorted_ant, self.niching_size)
                    niche.append(subniche)
                else:
                    subniche, ant = self.nearestdistance(sorted_ant[0, :], sorted_ant, self.niching_size)
                    sorted_ant = self.crowddistance(referpoint, ant)
                    niche.append(subniche)
        else:
            sorted_ant = self.crowddistance(referpoint, self.position)
            for i in range(int(len(self.position) / self.niching_size) + 1):
                if len(sorted_ant) <= self.niching_size:
                    subniche = self.nearestdistance(sorted_ant[0, :], sorted_ant, self.niching_size)
                    niche.append(subniche)
                else:
                    subniche, ant = self.nearestdistance(sorted_ant[0, :], sorted_ant, self.niching_size)
                    sorted_ant = self.crowddistance(referpoint, ant)
                    niche.append(subniche)
        Solutions = niche[0]
        cluster = list()
        for i in range(1, len(niche)):
            Solutions = np.concatenate((Solutions, niche[i]), axis=0)
        for j in range(len(niche)):
            for k in range(len(niche[j])):
                cluster.append(int(j))
        cluster = pd.DataFrame(cluster, columns=['c'])
        position = pd.DataFrame(Solutions)
        fitness = position.apply(Algorithm().costfunction, axis=1).astype('float')
        fitness = pd.DataFrame(fitness, columns=['f'])
        result = pd.concat([cluster, fitness, position], axis=1)
        return cluster, fitness, position, result

    def crowddistance(self, individual, popu):
        r"""
        :param individual: An individual
        :param popu: A population
        :return: sorted result by euclidean distance
        """
        distance = 0.0
        ant_distance = np.zeros(len(popu))
        for i in range(len(popu)):
            for j in range(self.benchmark.dimension):
                distance += (popu[i, j] - individual[0, j]) ** 2
            ant_distance[i] = np.sqrt(distance)
        connection = np.concatenate((ant_distance[:, np.newaxis], popu), axis=1)
        connection = connection[connection[:, 0].argsort()]
        sorted_connection = connection[:, 1:self.benchmark.dimension + 1]
        return sorted_connection

    def nearestdistance(self, individual, popu, M):
        r"""
        :param individual: An individual
        :param popu: A population
        :param M: type size
        :return: sorted individuals by distance
        """
        distance = 0.0
        indivi_distance = np.zeros(len(popu))
        for i in range(len(popu)):
            for j in range(self.benchmark.dimension):
                distance += (individual[j] - popu[i, j]) ** 2
            indivi_distance[i] = np.sqrt(distance)
        indivi_connect = np.concatenate((indivi_distance[:, np.newaxis], popu), axis=1)
        indivi_connect = indivi_connect[indivi_connect[:, 0].argsort()]
        sorted_connection = indivi_connect[:, 1:self.benchmark.dimension + 1]
        if len(popu) <= M:
            result = sorted_connection
            return result
        else:
            result = sorted_connection[0:M, :]
            rest_popu = sorted_connection[M:, :]
            return result, rest_popu
